/*
 * Copyright (c) 1991-1995 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Network Research
 *	Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: cbq.cc,v 1.4 95/10/11 16:53:44 mccanne Exp $ (LBL)";
#endif

/*
 * Here is a sample set of RM parameters, from one simulation:
 * 
 * rm [ idletype=0 equiv_plot=0 weighted=1 maxpkt=1000 delayqueue=1 use_depth=1]
 * 
 * We have made no attempt to produce efficient code; our goal has simply
 * been to write code that correctly implements the many algorithms
 * that we have been experimenting with.  This code has NO
 * relationship to production code that would be used to implement any of 
 * these resource management algorithms in routers.
 * 
 * The code is hard-to-understand in part because one of the goals
 * of our simulations is to compare a number of different algorithms. 
 *   The parameter "use_depth" chooses one of three different
 * link-sharing algorithms:  use_depth=0 for Ancestor-Only link-sharing,
 * use_depth=1 for Top-Level link-sharing, and use_depth=2 for Formal
 * link-sharing.  
 *   The parameter "idletype" picks the method for computing "idle"; 
 * idletype=0 for the original method, idletype=1 for a modified method
 * that lets you more accurately estimate the actual bandwidth used
 * by the class.
 *   The parameter "weighted" picks the round-robin algorithm;
 * weighted=0 for packet-by-packet round-robin, and weighted=1 for
 * weighted round-robin.  (The implementation of weighted round-robin
 * could use a great deal of improvement.)
 *   The parameter "delayqueue" picks the implementation of the
 * overlimit action;  delayqueue=0 simply uses the "undertime"
 * parameter, and delayqueue=1 removes overlimit classes from the
 * active list to the delay queue.
 *   The parameter "equiv_plot=1", along with the procedures
 * rm_equivplot(), rm_equiv(), admitted(), and admit_conn(), is
 * used for experimenting with a particular admissions control procedure.
 * 
 * Here is a sample set of parameters for a particular class:
 * 
 * class Audio [parent=Top borrow=Top queue=ubq pri=1 depth=0 allotment=10%
 *    overlimit=delay maxidle=4.0ms minidle=-0.2ms extradelay=0ms
 *    plot-file=none epsilon=0.01 weight=0.01 admit=0 ] 
 * 
 *   The parameters "parent" and "borrow" indicate the parent class
 * in the class structure, and the class to borrow bandwidth from.
 * Generally these two parameters will be the same.
 *   The parameter "queue" indicstes the queueing discipline.
 *   The parameter "depth" indicates the "depth" of the class in
 * the class structure, used by Top-Level link-sharing.  Leaf classes have
 * "depth=0".
 *   The parameter "allotment" indicates the link-sharing bandwidth
 * allocated to the class, as a fraction of the link bandwidth.
 *   The parameter "overlimit" indicates the overlimit action for the
 * class.
 *   Parameters "maxidle", "minidle", and "extradelay" are used in
 * calculating the bandwidth used by the class.  Guidelines for
 * setting these parameters are in the file ~floyd/paper/rm/rm.tex.
 *   The parameter "admit=1" indicates a class with admissions
 * control.  The parameters "epsilon" and "weight" are used by the
 * admissions control procedure.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: cbq.cc,v 1.4 95/10/11 16:53:44 mccanne Exp $ (LBL)";
#endif

#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include "config.h"
#include "real.h"
#include "packet.h"
#include "class.h"
#include "at.h"
#include "bind.h"

#define MAXDEPTH 32
#define POWEROFTWO 16		/* for calculating avgidle */
#define INTERVAL 0.05		/* time interval for equi. cap. calculations */

#include "link.h"

struct statlist {
        int pkts;       /* no. of packets transmitted by class */
        int bytes;      /* no. of bytes transmitted by class */
	double mean;	/* mean class bandwidth in bytes/sec */
	double var;	/* variance of class bandwidth */
	double cap;	/* equivalent capacity */
	int conns;	/* number of admitted connections */
};

struct idlist {
	struct idlist *next;
	char *id;
};

class CBQueue;

struct Class {
	Link* queue;
	CBQueue* master;
	double last;
	double undertime;
	double avgidle;
	statlist stats;		/* for the admission control procedure */

	Class* peer;
	Class* parent;
	Class* borrow;
	Class* equiv;		/* next class with admissions controls */
	int pri;		/* priority */
	int depth;		/* "depth" of class in class tree */
	void (*overlimit)(Class*);
	double maxidle;
	double minidle;
	double extradelay;
	double maxrate;		/* bytes per second allowed this class */
	int admit;      	/* 1 for admissions controls, 0 for not */
	double epsilon;		/* for equivalent capacity */
	double weight;		/* for computing average class throughput */
	int bytes_alloc;	/* num. of bytes allocated for current round */
	int delayed;		/* 1 for an already-delayed class */
        char *plot_fname;
	FILE *plotfile;

	/*
	 * Temporary things while constructing data structures.
	 */
	char *name;
	char *pname;
	char *bname;
	char *qname;
	char *oname;
	double allotment;
	Class *next;
	idlist *members;
	int lineno;
	int mark;
};

class CBQueue : public Link {
 public:	
	REDQueue();
 protected:
	Class* active_[MAXPRIO];	/* active class at each level */
	Class* equivcl_;   /* class with admissions control */
	Class* cc_;	/* current class */
	Class* borrowed_;	/* the class last borrowed from */
	Class** classtab_;	/* class of newly-arrived pkts */
	Class* list_;	/* beginning of list of classes */
	int idletype_;		/* 0 for idle as a difference, */
				/* 1 for idle as a fraction */
	int equiv_plot_;		/* 1 to plot equivalent capacity */
	int weighted_;		/* 1 for weighted RR, 0 otherwise */
	int delayqueue_;		/* 1 for delay queues, 0 for not */
	int maxpkt_;		/* max packet size in bytes */
	int deficit_;		/* For the weighted RR scheduling. 
				 * If deficit==1 after one priority level
				 * of scheduling, then change deficit=2,
				 * and repeat that priority level round.
				 */
	int num_[MAXPRIO];	/* number of classes at each priority level */
	int depth_;		/* max depth at which borrowing is allowed */
	int use_depth_;		/* 1 to restrict borrowing, 0 otherwise */
				/* 2 for idealized link-sharing guidelines */
	double alloc_[MAXPRIO];	/* total allocation at each level, for WRR */
	double M_[MAXPRIO];	/* for weighted RR, using num[] and alloc[] */
	double last_;		/* last time to compute equivalent capacities */
	Class* admitlist_[MAXIDS];	
				/* for each conn. id, the class to which it 
				   has been admitted. */
};

CBQueue::CBQueue()
{
	Tcl& tcl = Tcl::instance();
	idletype_ = tcl.rdi("ns_cbq", "idletype");
	equiv_plot_ = tcl.rdi("ns_cbq", "equiv_plot");
	weighted_ = tcl.rdi("ns_cbq", "weighted");
	delayqueue_ = tcl.rdi("ns_cbq", "delayqueue");
	use_depth_ = tcl.rdi("ns_cbq", "use_depth");

	depth_ = MAXDEPTH;
	for (int i = 0; i < MAXPRIO; ++i) {
		num_[i] = 0;
		alloc_[i] = 0.0;
		M_[i] = 0.0;
	}
	for (i=0; i < MAXIDS; ++i)
		admitlist_[i] = 0;

	last_ = 0.0;
	/*XXX all other */
}

/*XXX overlimt action -- need "this*/
void CBQueue::class_drop_packet(Class* cl)
{
	Link* link = cl->queue;
	Packet* pkt = cl->queue->deque();
	if (pkt != 0) {
        	/*
          	* Don't count this as a packet sent by the gateway!
        	*/
		/*XXX*/
#ifdef notdef
		--qd->qd_delay[pkt->cid].a_nelm;
#endif
		--qlen_;
		drop(pkt);
	}
}

/* 
 * Class_delay_action removes class from the active class list,
 *    if "delayqueue" is 1.  Otherwise it just sets cl->undertime.
*/
void
class_delay_action(register struct class *cl)
{
	register struct class **cp = &cl->stp->active[cl->pri];
	register struct class *p, *head, *orig_cl;
	double now = realtime();
	double delay;

	delay = cl->undertime - now + cl->extradelay;
        if ( delay > 0 && cl->stp->delayqueue==1) {
	  orig_cl = cl;
	  head = *cp;
	  if (cl == cl->peer) {
		/*
		 * Only member in active list.
		 */
		*cp = 0;
          } else {
            if (cl == head) {
                for (p = head->peer; cl != p->peer; p = p->peer)
                       if (p == head)
                                fatal("class_delay_action: link not found");
	  	p->peer = cl->peer;
		*cp = cl->peer;
	    }
            else {
		for (p = head; cl != p->peer; p = p->peer)
                        if (p->peer == head)
                                fatal("class_delay_action: class not found");
        	p->peer = cl->peer;
	    }
          }
          classtimer(orig_cl, (long)(1e6 * delay), orig_cl->ifp);
	}
	else if (delay>0 && cl->stp->delayqueue==0 && cl->delayed==0 ) {
	  cl->undertime += cl->extradelay;
	  cl->undertime -= (1-POWEROFTWO)*cl->avgidle;
	  cl->delayed = 1;
	}

}

/*
 * Ancestor returns 1 if class cl is an ancestor of class p. 
 * This is used only for use_depth_ ==2.
 */
int CBQueue::ancestor(Class* cl, Class* p)
{
	for (Class* b = p->borrow; b != 0; p = b)
		if (cl == b)
			return (1);
	return (0);
}

/* Satisfied returns 1 if the class is satisfied.
 * This is used only for stp->use_depth==2.
 */
int CBQueue::satisfied(Class* cl)
{
	Class* p;
	double now = Scheduler::instance.clock();
	if (cl->undertime > now)
		return (1);
	else if (cl->depth == 0) {
		if (cl->queue->qlen_ > 0)
			return (0);
		else
			return (1);
	} else {
		for (p = list_; p != 0; p = p->next) {
			if (p->depth == 0 && ancestor(cl, p)) {
				if (p->queue->qlen_ > 0)
					return (0);
			}
		}
	}
	return (1);
}

/*
 * Ok_to_borrow returns 1 if it is ok for a descendant to borrow
 * from this underlimit interior class cl. 
 * This is used only for stp->use_depth==2.
 */
int CBQueue::ok_to_borrow(Class* cl)
{
	for (Class* p = list_; p != 0; p = p->next)
		if (ancestor(cl, p) && !satisfied(p))
			return (0);
	return (1);
}

/*
 * Class_under_limit tests whether a class is either underlimit
 * or able to borrow.  The maximum depth at which the class is allowed
 * to borrow is "stp->depth".
 */
int CBQueue::class_under_limit(Class* cl, double now)
{
	if (cl->undertime < now)
		cl->delayed = 0;
	else {
		while (cl->undertime > now || 
		       (use_depth_ == 2 && !ok_to_borrow(cl))) {
			cl = cl->borrow;
			if (cl == 0) 
				return (0);
			if (use_depth_ == 1) {
				if (cl->depth > depth_)
					return (0);
			}
		}
		borrowed_ = cl;
		return (1);
	}
}

/*
 * Class_restart inserts a class that has been removed from the list
 * of active classes.  This is only needed when "delayqueue" is 1.
 */
void CBQueue::class_restart(Class* cl)
{
	Class** active = &active_[cl->pri];
	Class* peer = *active;
	if (peer != 0) {
		cl->peer = peer->peer;
		peer->peer = cl;
	} else 
		cl->peer = cl;
	*active = cl;
}

void CBQueue::plot(Class* cl, double idle)
{
        double t = Schedular.instance().clock();
        fprintf(cl->plotfile, "%g avg %g\n", t, cl->avgidle);
	fprintf(cl->plotfile, "%g idl %g\n", t, idle);
}

void CBQueue::equivplot(Class* cl, double thru)
{
        double t = Schedular.instance().clock();
	fprintf(cl->plotfile, "%g mea %g\n", t, cl->stats.mean);
        fprintf(cl->plotfile, "%g var %g\n", t, cl->stats.var);
        fprintf(cl->plotfile, "%g cap %g\n", t, cl->stats.cap);
	fprintf(cl->plotfile, "%g thr %g\n", t, thru);
}

/*
 * Update_class_util calculates "idle" and "avgidle", and updates "undertime",
 * immediately after a packet is sent from a class.
 * Update_class_util also updates the "depth" parameter:  if this class
 * could send another packet, borrowing from depth at most 
 * "borrowed->depth", then "stp->depth" is set to
 * at most "borrowed->depth".  
 */
void CBQueue::update_class_util(Class* cl, double speed, double now,
				int pktsize)
{
	Class* cl0 = cl;
	Link* link = cl->link;

	double txt = speed * pktsize;
	now += txt;
	do {
		idle = (now - cl->last) - (pktsize / cl->maxrate);
		if (cl->master->idletype_ != 0) 
			idle -= 1.;
		avgidle = cl->avgidle;
		avgidle += (idle - avgidle) / POWEROFTWO;
		if (avgidle > cl->maxidle)
			avgidle = cl->maxidle;
		cl->avgidle = avgidle;
		if (avgidle <= 0) {
			if (avgidle < cl->minidle) 
				avgidle = cl->minidle;
			cl->undertime = now + txt * (1 / cl->allotment - 1); 
			cl->undertime += (1-POWEROFTWO) * avgidle;
			if (cl->pri == MAXPRIO) 
				cl->undertime += cl->extradelay;
		}
		cl->last = now;
		if (cl->plotfile)
			plot(cl, idle);
	} while (cl = cl->parent);

	if (use_depth == 1 && depth_ >= borrowed_->depth) {
		if (link_->qlen_ <= 1 <= 1 || borrowed_->undertime > now)
			depth_ = MAXDEPTH;
		else 
			depth_ = borrowed_->depth;
	}
}

/*
 * Compute the average and the variance of the bandwidth, and the
 * equivalent capacity, for those classes with admissions controls.
 */
void CBQueue::compute_equiv()
{
	double now = Scheduler::instance().clock();
	double interval = now - stp->last;
	for (Class* cl = equivcl_; cl != 0; cl = cl->equiv) {
		double thru = cl->stats.bytes / interval;
		double diff = thru - cl->stats.mean;
		cl->stats.mean += cl->weight * diff;
		cl->stats.var += cl->weight / 4 * (diff * diff - cl->stats.var);
                if (cl->stats.var <= 0)
                    	cl->stats.cap = cl->stats.mean;
		else if (cl->stats.conns == 0)
			cl->stats.cap = 0.0;
		else {
			cl->stats.cap = cl->stats.mean +
				0.7 * sqrt(cl->stats.conns) *
				sqrt(log(1. / cl->epsilon)) *
				(cl->stats.mean/cl->stats.conns +
				 cl->stats.var/cl->stats.mean);
			cl->stats.bytes=0;
			cl->stats.pkts=0;
			if (equiv_plot == 1) 
				CBQueue::equivplot(cl, thru);
		}
		last_ = now;
}

/*
 * The heart of the scheduler, which decides which class
 * next gets to send a packet.  Highest priority first,
 * round-robin within priorities.
 *
 * For Weighted RR (weighted==1), each able-to-send class gets to send until
 * its byte allocation is exausted.  Thus "*active" is only changed after
 * a class is finished sending its allocation.
 *
 * If the scheduler finds no class that is underlimit or able to borrow, then
 * "firstclass" is the first class found that had a nonzero queue
 * and is allowed to borrow.  Class "firstclass" gets to send a packet.
 */
Packet* CBQueue::deque()
{
	double now = Scheduler::instance().clock();
	if (now - last_ > INTERVAL) 
		compute_equiv();
	deficit_ = 0;
	Class* first = 0;
	Class** active = active_;
	for (i = 0; ++i < MAXPRIO; ++active) {
		Class* cl = *active;
		int done = 0;
		/*
		 * Loop through twice for a priority level, if some class
		 * was unable to send a packet the first round because
		 * of the weighted round-robin mechanism.
		 * During the second loop at this level, deficit==2.
		 * (This second loop is not needed if for every class,
		 * "stp->M[cl->pri])" times "cl->allotment" is greater than
		 * the byte size for the largest packet in the class.)
		 */
		while (!done) {
			while (cl != 0) {
				Link* link = cl->link;
				if (weighted_==1 && deficit_ < 2 && cl->bytes_alloc <= 0) 
					cl->bytes_alloc += (int) (cl->allotment * stp->M[cl->pri]);
				if (link->qlen_ > 0) {
					if (first == 0 && cl->borrow != 0)
						first = cl;

					if (!class_under_limit(cl, now))
						(*cl->overlimit)(cl);
					else {
						if (weighted_ == 0 || cl->bytes_alloc > 0 ||
						    deficit_ > 1) {
							pkt = link->deque();
							/*XXX*/
							--qlen_;
							if (pkt != 0) {
								if (weighted_ == 1 && cl->bytes_alloc > 0)
									cl->bytes_alloc -= pkt->size;
								update_class_util(cl, stp->speed, now,
										  pkt->size);
								class_ = cl;
								if (weighted_ == 0 || cl->bytes_alloc <= 0)
									*active = cl->peer;
								else
									*active = cl;
								return (pkt);
							}
						} else
							deficit_ = 1;
					}
				}
				if (weighted_ == 1 && cl->bytes_alloc > 0)
					cl->bytes_alloc = 0;
				cl = cl->peer;
				if (cl == *active || *active == 0) 
					break;
			}
			if (weighted_ == 1 && deficit_ == 1) 
				deficit_ = 2;
			else
				done = 1;
		}
	}
	/*
	 * If no class is underlimit or able to borrow ...
	 */
	depth_ = MAXDEPTH;
	if (first != 0) {
		pkt = first->link->deque();
		if (pkt != 0) {
			class_ = first;
			active_[first->pri] = first->peer;
			return (pkt);
		}
	}
	class_ = 0;
	return (0);
}

/*
 * Add_conn adds a conversation id to the array of admitted connections.
 */
void CBQueue::add_conn(short cid, Class* cl)
{
	admitlist_[(int) cid] = cl;
}

/*
 * Del_conn deletes a conversation id from the array of admitted connections.
 */
void CBQueue::del_conn(short cid, Class* cl)
{
	admitlist_[(int) cid] = 0;
}

/*
 * Returns 1 if this connection has already been admitted to this class.
 */
int CBQueue::admitted(Packet* pkt, Class* cl)
{
#ifdef notyet
	if (admitlist_[(int) pkt->cid] == cl) 
		return (1);
	else
#endif
		return (0);
}

/*
 * Decide whether to admit connection pkt->cid to class cl.
 * For an ON/OFF connection, the additional variance would be
 * pkt->avgrate*(pkt->maxrate - pkt->avgrate).  We overestimate
 * the variance somewhat.
 */
int CBQueue::admit_conn(Packet* pkt, Class* cl)
{
#ifdef notdef
	if (cl->maxrate - cl->stats.cap > pkt->maxrate
	    && admitlist_[(int) pkt->cid]==0 ) {
		cl->stats.mean +=  pkt->avgrate;
		cl->stats.var += pkt->maxrate * pkt->maxrate;
		return (1);
	} else
#endif
		return (0);
}

/*
 * Map arriving packets to the appropriate class, and adds the 
 * arriving packet to the queue for that class.
 * If stp->use_depth==1, and that class could send a packet borrowing 
 * from depth at most d, then "stp->depth" is set to no more than d.
 * For new connections to classes with admissions controls, an
 * admissions control procedure has to be checked.
 */
void CBQueue::enque(Packet* pkt)
{
	double now = Scheduler::instance().clock();
	Class* cl = classify(pkt);
	if (cl == 0) {
		/*XXX error*/
		abort();
	}
	if (cl->admit == 1) {
		if (pkt->type == PSTART) { 
			/*
			 * Check for admissions.  
			 */
			if (admit_conn(pkt, cl)) {
				++cl->stats.conns;
				add_conn(pkt->cid, cl);
			}
		}
	}
	if (pkt->type == PSTOP && admitted(pkt, cl)) {
	  	if (cl->stats.conns > 0)
			--cl->stats.conns;
		del_conn(pkt->cid, cl);
	}
	if (!cl->admit || admitted(pkt, cl) || pkt->type == PSTOP) {
		cl->link->enque(&pkt);
		++qlen_;
		++cl->stats.pkts;
		cl->stats.bytes += pkt->size;
		if (use_depth_ == 1 && depth_ > 0) {
			if (cl->undertime < now) 
				depth_ = 0;
			else if (depth_ > 1 && cl->borrow != 0) {
				if (cl->borrow->undertime < now)
					depth_ = 1;
			}
		}
	} else
		drop(pkt);
}

typedef void (*olfn)(struct class *);

olfn
rm_olfunc(char *name)
{
	if (name == 0 || name == intern("drop"))
		return class_drop_packet;
	else if (name == intern("delay"))
		return class_delay_action;
}

struct class *
cl_parent(struct class *cl)
{
	return cl->parent;
}

struct class *
cl_borrow(struct class *cl)
{
	return cl->borrow;
}

void
cl_trav(struct class *p, struct class *(*parent)(struct class *))
{
	struct class *par;

	if (p->mark)
		return;
	p->mark = -1;
	par = parent(p);
	if (par != 0) {
		cl_trav(par, parent);
		p->mark = par->mark;
	}
}

/*
 * Check that a linkage forms a forest.
 */
void
cl_check(struct class *clist, struct class *(*parent)(struct class *))
{
	struct class *p;

	for (p = clist; p != 0; p = p->next)
		p->mark = (parent(p) == 0) ? 1 : 0;

	for (p = clist; p != 0; p = p->next)
		cl_trav(p, parent);
	for (p = clist; p != 0; p = p->next)
		if (p->mark <= 0) 
			error("(line %d): class set has cycle", clist->lineno);
}

/*
 * class options
 */
static struct at_options cloptions[] = {
	{ "pri", AT_INT, offsetof(struct class, pri) },
	{ "depth", AT_INT, offsetof(struct class, depth) },
        { "admit", AT_INT, offsetof(struct class, admit) },
	{ "allotment", AT_DOUBLE, offsetof(struct class, allotment) },
	{ "maxidle", AT_DOUBLE, offsetof(struct class, maxidle) },
	{ "minidle", AT_DOUBLE, offsetof(struct class, minidle) },
	{ "overlimit", AT_STRING, offsetof(struct class, oname) },
	{ "extradelay", AT_DOUBLE, offsetof(struct class, extradelay) },
	{ "parent", AT_NSTRING, offsetof(struct class, pname) },
	{ "borrow", AT_NSTRING, offsetof(struct class, bname) },
	{ "queue", AT_STRING, offsetof(struct class, qname) },
	{ "plot-file", AT_NSTRING, offsetof(struct class, plot_fname) },
        { "epsilon", AT_DOUBLE, offsetof(struct class, epsilon) },
        { "weight", AT_DOUBLE, offsetof(struct class, weight) },
	{ 0 }
};

/* For initialization.
 */
static void
rm_initclass(struct rmstate *stp, int id, double bps)
{
	struct class *list, *cl, **ctab, **active, *p;
	struct ifsw *isw;
	struct ifnet *ifp;
	struct idlist *m;
	int n, cid, i;
	char *cp;

	n = n_convs * sizeof(struct class *);
	ctab = (struct class **)malloc(n);
	bzero((char *)ctab, n);
	stp->classtab = ctab;
	bzero((char *)stp->active, sizeof(stp->active));
	active = stp->active;

	list = getclasslist(id, &ifp);
	stp->list = list;
	/*
	 * intialize attributes
	 */
	for (p = list; p != 0; p = p->next) {
		p->stp = stp;
		atl_init(p->nb, p->attr, cloptions, (char *)p);
                p->undertime = 0;
                p->last = 0;
                p->avgidle = 0;
		p->delayed = 0;
		p->equiv = 0;
		if (p->admit==1) {
			p->stats.pkts = 0;
			p->stats.bytes = 0;
			p->stats.mean = 0.0;
			p->stats.var = 0.0;
			p->stats.conns = 0;
		}
		if (stp->idletype==0) {
		  p->maxidle *= POWEROFTWO;
		  p->minidle *= POWEROFTWO;
		}
		p->ifp = ifp;
	        if (p->plot_fname != 0) {
	                FILE *f = fopen(p->plot_fname, "w");
	                if (f == 0)
	                        error("cannot open %s", p->plot_fname);
	                p->plotfile = f;
	        }
	}
	/*
	 * fix up borrowers and lenders
	 */
	for (p = list; p != 0; p = p->next) {
		/*
		 * Lookup queue proto.
		 */
		cp = p->qname;
		if (cp == 0)
			cp = intern("ubq");
		if (iflookup(cp, &p->qd.qd_sw) == 0)
			error("(line %d) class %s: protocol %s not found", 
			      p->lineno, p->name, cp);
		qdinit(&p->qd, 0.0);
		p->qd.qd_id = ifp->if_edge->re_id;

		/*
		 * Check that prioroty is in range.
		 */
		if (p->pri < 0 || p->pri > 32)
			error("(line %d): class %s: priority %d out of range",
			      p->lineno, p->name, p->pri);
		if (active[p->pri] != 0) {
			p->peer = active[p->pri]->peer;
			active[p->pri]->peer = p;
		} else {
			p->peer = p;
			active[p->pri] = p;
		}
		++stp->num[p->pri];
		stp->alloc[p->pri] += p->allotment;
		/*
		 * Set overlimit function.
		 */
		p->overlimit = rm_olfunc(p->oname);
		/*
		 * Compute maxrate from allotment.
		 */
		p->maxrate = p->allotment * bps;
		if (p->pname != 0) {
			cl = classlookup(list, p->pname);
			if (cl == 0)
				error("(line %d) parent %s of %s not found", 
				      p->lineno, p->pname, p->name);
			p->parent = cl;
		} else
			p->parent = 0;
		if (p->bname != 0) {
			cl = classlookup(list, p->bname);
			if (cl == 0)
				error("(line %d) lender %s of %s not found", 
				      p->lineno, p->pname, p->name);
			p->borrow = cl;
		} else
			p->borrow = 0;
	}
	/*
	 *  Make list of classes with admissions controls.
	 */
	stp->equivcl = 0;
	for (p = list; p != 0; p = p->next) {
	    if (p->admit==1) {
		p->equiv = stp->equivcl;
		stp->equivcl = p;
	    }
	}
	for (i=0; i<MAXPRIO; ++i) {
	  if (stp->alloc[i] > 0)
	    stp->M[i] = stp->num[i] * stp->maxpkt * 1.0/stp->alloc[i];
	  else stp->M[i] = 0;
	}
	/*
	 * Check that parent and borrow linkage each form a forest.
	 */
	cl_check(list, cl_parent);
	cl_check(list, cl_borrow);

	/*
	 * Set up class table to map conversation id to class struct.
	 */
	for (p = list; p != 0; p = p->next) {
		for (m = p->members; m != 0; m = m->next) {
			cid = getconnid(m->id);
			if (cid < 0)
	error("(line %d) class %s: cannot find conversation \"%s\"", 
			      p->lineno, p->name, m->id);
			ctab[cid] = p;
		}
	}
}

struct qdisc *
find_queue(struct qdisc *qd, int pr_cid, struct ifnet *ifp)
{
  struct rmstate *stp = STATE(qd);
  struct class **active = stp->active;
  struct class *p;
  struct idlist *m;
  struct class *list;

  list = getclasslist(qd->qd_id, &ifp);
  for (p = list; p != 0; p = p->next) {
        for (m = p->members; m != 0; m = m->next) {
            if (pr_cid == getconnid(m->id) ) {
              return &p->qd; break;
	     }
        }
    }
}
