/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: command.cc,v 1.11 95/09/29 15:08:40 mccanne Exp $ (LBL)";
#endif

#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "Tcl.h"
#include "command.h"
#include "trace.h"

CommandHandler* CommandTable::hashtab_[CMD_HASH_SIZE];

/* "hashpjw" from Aho, Sethi & Ullman, p.436 */
int CommandTable::hash(const char* cp)
{
	register unsigned int h = 0, g;

	while (*cp) {
		h = (h << 4) + *cp++;
		if (g = h & 0xf0000000) {
			h = h ^ (g >> 24);
			h = h ^ g;
		}
	}
	return (h & (CMD_HASH_SIZE-1));
}

int CommandTable::command(ClientData, Tcl_Interp*, int argc, char** argv)
{
	if (argc < 2) {
		/*XXX*/
		abort();
	}
	Tcl& tcl = Tcl::instance();
	const char* cmd = argv[1];
	int h = hash(cmd);
	for (;;) {
		CommandHandler* p = hashtab_[h];
		if (p == 0) {
			tcl.resultf("%s %s: no such command",
				    argv[0], argv[1]);
			return (TCL_ERROR);
		}
		if (strcmp(p->name(), cmd) == 0)
			return (p->command(argc, argv));

		h = (h + 1) & (CMD_HASH_SIZE-1);
	}
}

void CommandTable::enter(CommandHandler* p)
{
	int h = hash(p->name());
	while (hashtab_[h] != 0)
		h = (h + 1) & (CMD_HASH_SIZE-1);
	hashtab_[h] = p;
}

CommandHandler::CommandHandler(const char* name) : name_(name)
{
	CommandTable::enter(this);
}

class Var {
 public:
	Var(const char* name, int readonly);
	virtual ~Var();
	Var* next_;
	inline const char* name() const { return (name_); }
	virtual void set(const char*) = 0;
	virtual const char* get() = 0;
	inline int readonly() const { return (readonly_); }
 protected:
	char* name_;
	int readonly_;
	/*XXX make this static? */
	char wrk_[32];
};

class VarReal : public Var {
 public:
	VarReal(const char* name, int readonly, double* val);
	virtual void set(const char*);
	virtual const char* get();
 protected:
	double* val_;
};

class VarBandwidth : public VarReal {
 public:
	VarBandwidth(const char* name, int readonly, double* val)
		: VarReal(name, readonly, val) { }
	virtual void set(const char*);
};

class VarTime : public VarReal {
 public:
	VarTime(const char* name, int readonly, double* val)
		: VarReal(name, readonly, val) { }
	virtual void set(const char*);
};

class VarInt : public Var {
 public:
	VarInt(const char* name, int readonly, int* val);
	virtual void set(const char*);
	virtual const char* get();
 protected:
	int* val_;
};

class VarBool : public VarInt {
 public:
	VarBool(const char* name, int readonly, int* val);
	virtual void set(const char*);
};

Object* Object::all_;
int Object::nid_;

Object::Object(int prefix) : var_(0), trace_(0)
{
	id_ = nid_++;
	sprintf(proc_, "_%c%d", prefix, id_);
	Tcl& t = Tcl::instance();
	t.CreateCommand(proc_, Object::command, (ClientData)this);
	next_ = all_;
	all_ = this;
}

Object::~Object()
{
	Object** o; 

	Tcl& t = Tcl::instance();
	t.DeleteCommand(proc_);

	for (Var* p = var_; p != 0; ) {
		Var* n = p->next_;
		delete p;
		p = n;
	}
	for (*o = all_; *o != this; o = &(*o)->next_)
		;
	*o = (*o)->next_;
}

void Object::list(int prefix, char* cp)
{
	for (Object* p = all_; p != 0; p = p->next_) {
		if (p->proc_[1] == prefix) {
			strcpy(cp, p->proc());
			cp += strlen(cp);
			*cp++ = ' ';
		}
	}
	*cp = 0;
}

void Object::reset_all()
{
	for (Object* p = all_; p != 0; p = p->next_)
		p->reset();
}

Object* Object::lookup(const char* name)
{
	for (Object* p = all_; p != 0; p = p->next_)
		if (strcmp(p->proc_, name) == 0)
			return (p);
	return (0);
}

void Object::reset()
{
}

void Object::insert(Var* p)
{
	p->next_ = var_;
	var_ = p;
}

int Object::command(ClientData cd, Tcl_Interp*, int argc, char **argv)
{
	Object* o = (Object*)cd;
	if (argc == 4 && strcmp(argv[1], "set") == 0)
		return (o->set(argv[2], argv[3]));
	else if (argc == 3) {
		if (strcmp(argv[1], "set") == 0 ||
		     strcmp(argv[1], "get") == 0)
			return (o->get(argv[2]));
		if (strcmp(argv[1], "trace") == 0) {
			o->trace_ = Trace::lookup(argv[2]);
			if (o->trace_ == 0) {
				Tcl& tcl = Tcl::instance();
				tcl.resultf("ns object trace: no trace %s",
					    argv[2]);
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
	}
	return (o->command(argc, argv));
}

Var* Object::varlookup(const char* var) const
{
	for (Var* p = var_; p != 0; p = p->next_)
		if (strcmp(p->name(), var) == 0)
			return (p);
	Tcl& tcl = Tcl::instance();
	tcl.resultf("ns: no such var: %s", var);
	return (0);
}

int Object::get(const char* var)
{
	Var* v = varlookup(var);
	if (v == 0)
		return (TCL_ERROR);
	Tcl::instance().result(v->get());
	return (TCL_OK);
}

int Object::set(const char* var, const char* val)
{
	Var* v = varlookup(var);
	if (v == 0)
		return (TCL_ERROR);
	if (v->readonly()) {
		Tcl::instance().resultf("ns: %s is read-only", var);
		return (TCL_ERROR);
	}
	v->set(val);
	return (TCL_OK);
}

int Object::command(int argc, const char*const* argv)
{
	/*XXX unknown command*/
	return (TCL_ERROR);
}

void Object::link_real(const char* a, const char* name,
		       double* val, int readOnly)
{
	VarReal* v = new VarReal(name, readOnly, val);
	if (a != 0)
		v->set(Tcl::instance().rds(a, name));
	insert(v);
}

void Object::link_bw(const char* a, const char* name,
		       double* val, int readOnly)
{
	VarBandwidth* v = new VarBandwidth(name, readOnly, val);
	if (a != 0)
		v->set(Tcl::instance().rds(a, name));
	insert(v);
}

void Object::link_time(const char* a, const char* name,
		       double* val, int readOnly)
{
	VarTime* v = new VarTime(name, readOnly, val);
	if (a != 0)
		v->set(Tcl::instance().rds(a, name));
	insert(v);
}

void Object::link_int(const char* a, const char* name,
		      int* val, int readOnly)
{
	VarInt* v = new VarInt(name, readOnly, val);
	if (a != 0)
		v->set(Tcl::instance().rds(a, name));
	insert(v);
}

void Object::link_bool(const char* a, const char* name,
		      int* val, int readOnly)
{
	VarBool* v = new VarBool(name, readOnly, val);
	if (a != 0)
		v->set(Tcl::instance().rds(a, name));
	insert(v);
}

double
bw_atof(const char* s) 
{
	char wrk[32];
	char* cp = wrk;
	while (isdigit(*s) || *s == 'e' || *s == '+' ||
	       *s == '-' || *s == '.')
		*cp++ = *s++;
	*cp = 0;
	double v = atof(wrk);
	switch (s[0]) {
	case 'k':
	case 'K':
		v *= 1e3;
		break;
	case 'm':
	case 'M':
		v *= 1e6;
		break;
	}
	if (s[1] == 'B')
		v *= 8;

	return (v);
}

double
time_atof(const char* s)
{
	char wrk[32];
	char* cp = wrk;
	while (isdigit(*s) || *s == 'e' || *s == '+' || *s == '-' || *s == '.')
		*cp++ = *s++;
	*cp = 0;
	double v = atof(wrk);
	switch (*s) {
	case 'm':
		v *= 1e-3;
		break;
	case 'u':
		v *= 1e-6;
		break;
	case 'n':
		v *= 1e-9;
		break;
	case 'p':
		v *= 1e-12;
		break;
	}
	return (v);
}

Var::Var(const char* name, int readonly) : readonly_(readonly)
{
	int n = strlen(name) + 1;
	name_ = new char[n];
	strcpy(name_, name);
}

Var::~Var()
{
	delete name_;
}

VarReal::VarReal(const char* name, int readonly, double* val)
	: Var(name, readonly), val_(val)
{
}

const char* VarReal::get()
{
	sprintf(wrk_, "%g", *val_);
	return (wrk_);
}

void VarReal::set(const char* s)
{
	*val_ = atof(s);
}

void VarBandwidth::set(const char* s)
{
	*val_ = bw_atof(s);
}

void VarTime::set(const char* s)
{
	*val_ = time_atof(s);
}

VarInt::VarInt(const char* name, int readonly, int* val)
	: Var(name, readonly), val_(val)
{
}

const char* VarInt::get()
{
	sprintf(wrk_, "%d", *val_);
	return (wrk_);
}

void VarInt::set(const char* s)
{
	*val_ = atoi(s);
}

VarBool::VarBool(const char* name, int readonly, int* val)
	: VarInt(name, readonly, val)
{
}

void VarBool::set(const char* s)
{
	if (isdigit(*s))
		*val_ = atoi(s);
	switch (*s) {
	case 't':
	case 'T':
		*val_ =  1;
		return;
	}
	*val_ = 0;
}
