/*
 * Copyright (c) 1990-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: link.cc,v 1.14 95/09/29 15:10:47 mccanne Exp $ (LBL)";
#endif

#include <stdlib.h>
#include <math.h>
#include <sys/types.h>

#include "link.h"
#include "packet.h"
#include "node.h"
#include "command.h"
#include "trace.h"

LinkMatcher* LinkMatcher::all_;

LinkMatcher::LinkMatcher()
{
	next_ = all_;
	all_ = this;
}

Link* LinkMatcher::lookup(const char* type)
{
	for (LinkMatcher* p = all_; p != 0; p = p->next_) {
		Link* link = p->match(type);
		if (link != 0)
			return (link);
	}
	return (0);
}

/*
 * tcl command interface
 */
static class LinkCommand : public CommandHandler {
public:
	LinkCommand(const char* name) : CommandHandler(name) { }
        virtual int command(int argc, const char*const* argv);
} link_cmd("link");

int LinkCommand::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if (argc == 2) {
		/* return list of all links */
		/*XXX buffer could overflow*/
		char* bp = tcl.buffer();
		Object::list('l', bp);
		tcl.result(bp);
		return (TCL_OK);
	}
	if (argc < 4) {
		tcl.resultf("ns: arg mismatch: %s", argv[1]);
		return (TCL_ERROR);
	}
	Node* src = Node::lookup(argv[2]);
	if (src == 0) {
		tcl.resultf("ns link: no such node: %s", argv[2]);
		return (TCL_ERROR);
	}
	Node* dst = Node::lookup(argv[3]);
	if (dst == 0) {
		tcl.resultf("ns link: no such node: %s", argv[3]);
		return (TCL_ERROR);
	}
	if (argc == 4) {
		/* lookup link */
		for (Link* p = src->links(); p != 0; p = p->next_) {
			if (p->neighbor() == dst) {
				tcl.result(p->proc());
				return (TCL_OK);
			}
		}
		tcl.resultf("ns: unknown link (%s,%s)", argv[2], argv[3]);
		return (TCL_ERROR);
	}
	
	Link* link = LinkMatcher::lookup(argv[4]);
	if (link == 0) {
		tcl.resultf("ns: unkown queue type: %s");
		return (TCL_ERROR);
	}

	link->neighbor(dst);
	src->add_link(link);

	tcl.result(link->proc());
	return (TCL_OK);
}

Queue::Queue() : tail_(&head_), head_(0)
{
}

void Queue::enque(Packet *p)
{
	p->next_ = 0;
	*tail_ = p;
	tail_ = &p->next_;
	*tail_ = 0;
}

/*
 * remove the nth (0...N-1) packet from the queue and return it
 */
Packet* Queue::deque(int n)
{
	Packet** p = &head_;

	while (--n >= 0)
		p = &(*p)->next_;
	Packet* pkt = *p;
	*p = pkt->next_;
	if (tail_ == &pkt->next_)
		tail_ = p;

	return (pkt);
}

Link::Link() : Object('l'), busy_(0), neighbor_(0), qnp_(0), qnb_(0),
	qtrace_(0)
{
	Tcl& tcl = Tcl::instance();
	link_bw("ns_link", "bandwidth", &bandwidth_, 0);
	link_time("ns_link", "delay", &delay_, 0);
	link_int("ns_link", "queue-limit", &qlim_, 0);

	link_int(0, "queue-length", &qnp_, 1);
	link_int(0, "queue-size", &qnb_, 1);

	ncs_ = 10;
	cs_ = new classStat[ncs_];
}

void Link::grow(int m)
{
	int i;
	int n = ncs_ << 1;
	while (n < m)
		n <<= 1;

	classStat* p = new classStat[n];
	for (i = 0; i < ncs_; ++i)
		p[i] = cs_[i];

	for (; i < n; ++i) {
		p[i].np = 0;
		p[i].nb = 0;
		p[i].nd = 0;
	}

	delete cs_;
	cs_ = p;
}

/*
 * Called every time the queue size changes.
 * We integrate the queue size and keep.
 */
void Link::update(const Packet* pkt)
{
	int c = pkt->class_;
	check_class(c);
	classStat& cs = cs_[c];
	++cs.np;
	cs.nb += pkt->size_;

	/* sum up the queue size (in both packets and bytes) */
	cs.sb.newpt(qnb_);
	cs.sp.newpt(qnp_);
	/* integrate the queue size (in both packets and bytes) */
	double now = Scheduler::instance().clock();
	ib_.newpt(now, qnb_);
	ip_.newpt(now, qnp_);
	iu_.newpt(now, (qnp_ != 0) ? 1. : 0.);
}

/*
 * Called every time we drop a packet.
 */
void Link::update_drop(const Packet* pkt)
{
	int c = pkt->class_;
	check_class(c);
	classStat& cs = cs_[c];
	++cs.nd;

	/* sum up the queue size (in both packets and bytes) */
	cs.sb.newpt(qnb_);
	cs.sp.newpt(qnp_);
	/* integrate the queue size (in both packets and bytes) */
	double now = Scheduler::instance().clock();
	ib_.newpt(now, qnb_);
	ip_.newpt(now, qnp_);
	iu_.newpt(now, (qnp_ != 0) ? 1. : 0.);
}

void Link::send()
{
	Packet* pkt = deque();
	if (pkt != 0) {
		double txtime = pkt->size_ * 8. / bandwidth_;
		Scheduler& s = Scheduler::instance();
		s.schedule(neighbor_, pkt, txtime + delay_);
		s.schedule(this, &intr_, txtime);
		busy_ = 1;
		busytime_ += txtime;
		if (trace_ != 0)
			trace_->hop(src_, dst_, pkt);
	}
}

/*
 * The only way this handler is called is from 
 * a "transmission interrupt".
 */
void Link::handle(Event* e)
{
	/*XXX*/
	if (e != &intr_)
		abort();

	busy_ = 0;
	send();
}

/*
 * Accept a packet for transmission on this link.
 * The node class calls this method.
 */
void Link::send(Packet* pkt)
{
#ifdef notyet
	if (!busy_ && qnp_ == 0) {
		/* no queue - just transmit the packet */
		if_sc(ifp, pkt);
		return;
	}
#endif
	enque(pkt);
	if (!busy_)
		send();
}

void Link::plot()
{
	double now = Scheduler::instance().clock();
	Trace* t = qtrace_;
	sprintf(t->buffer(), "q %g %d", now, qnp_);
	t->dump();
	sprintf(t->buffer(), "b %g %d", now, qnb_);
	t->dump();
}

/*
 * $link stat $class $which
 * $link integral qsize
 * $link integral qlen
 * $link integral util
 * $line queue-trace $trace
 */
int Link::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if (argc == 2) {
		if (strcmp(argv[1], "reset") == 0) {
			reset();
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(argv[1], "integral") == 0) {
			double now = Scheduler::instance().clock();
			if (strcmp(argv[2], "util") == 0) {
				iu_.update(now);
				tcl.resultf("%g", iu_.sum());
				return (TCL_OK);
			}
			if (strcmp(argv[2], "qsize") == 0) {
				ib_.update(now);
				tcl.resultf("%g", ib_.sum());
				return (TCL_OK);
			}
			if (strcmp(argv[2], "qlen") == 0) {
				ip_.update(now);
				tcl.resultf("%g", ip_.sum());
				return (TCL_OK);
			}
		}
		if (strcmp(argv[1], "queue-trace") == 0) {
			qtrace_ = Trace::lookup(argv[2]);
			if (qtrace_ == 0) {
				Tcl& tcl = Tcl::instance();
				tcl.resultf("link queue-trace: no trace %s",
					    argv[2]);
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
	} else if (argc == 4) {
		if (strcmp(argv[1], "stat") == 0) {
			int c = atoi(argv[2]);
			if (c >= ncs_) {
				tcl.result("0");
				return (TCL_OK);
			}
			classStat& cs = cs_[c];
			const char* s = argv[3];
			if (strcmp(s, "packets") == 0) {
				tcl.resultf("%d", cs.np);
				return (TCL_OK);
			}
			if (strcmp(s, "bytes") == 0) {
				tcl.resultf("%d", cs.nb);
				return (TCL_OK);
			}
			if (strcmp(s, "drops") == 0) {
				tcl.resultf("%d", cs.nd);
				return (TCL_OK);
			}
			/* average qlen seen by this class on arrival */
			if (strcmp(s, "mean-qlen") == 0) {
				tcl.resultf("%g", cs.sp.mean());
				return (TCL_OK);
			}
			/* average q size seen by this class on arrival */
			if (strcmp(s, "mean-qsize") == 0) {
				tcl.resultf("%g", cs.sb.mean());
				return (TCL_OK);
			}
		}
	}
	tcl.resultf("ns: no such link command: %s", argv[1]);
	return (TCL_ERROR);
}

void Link::reset()
{
	for (int i = 0; i < ncs_; ++i) {
		cs_[i].np = 0;
		cs_[i].nb = 0;
		cs_[i].nd = 0;
		cs_[i].sb.reset();
		cs_[i].sp.reset();
	}
	double now = Scheduler::instance().clock();
	iu_.reset(now);
	ib_.reset(now);
	ip_.reset(now);
}
