/*
 * Copyright (c) 1993 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) $Header: link.h,v 1.8 95/07/31 15:48:52 mccanne Exp $ (LBL)
 */

#ifndef tcpsim_link_h
#define tcpsim_link_h

#include "scheduler.h"
#include "command.h"
#include "packet.h"
#include "integrator.h"
#include "node.h"
#include "trace.h"

struct classStat {
	int np;			/* number of packets */
	int nb;			/* number of bytes */
	int nd;			/* number of drops */
	Sigma sb;		/* bytes */
	Sigma sp;		/* packets */
};

class Queue {
 public:
	Queue();
	void enque(Packet*);
	Packet* deque(int n);
 protected:
	Packet* head_;
	Packet** tail_;
};

/* XXX
 * The reprentation of an output link.  A node can have multiple output
 * lines (i.e., a router) or a single line (i.e., a source).  We maintain
 * the invariant that this structure is read-only to all modules except 
 * link.c (but the node field is writable only by the nodes).
 */
class Link : public Handler, public Object {
 public:
	void send(Packet*);
	void handle(Event*);
	int command(int argc, const char*const* argv);
	/* linked list grounded in node */
	Link* next_;
	inline Node* neighbor() const { return (neighbor_); }
	inline void neighbor(Node* p) { neighbor_ = p; dst_ = p->addr(); }
	inline void src(int s) { src_ = s; }
	inline int src() const { return (src_); }
	inline int dst() const { return (dst_); }
 protected:
	Link();
	virtual void reset();
	void update(const Packet*);
	void update_drop(const Packet*);
	void send();
	
	int src_;
	int dst_;

	int qnp_;		/* no. pkts currently in queue */
	int qnb_;		/* no. bytes currently in queue */
	int qlim_;		/* maximum allowed pkts in queue */
	virtual void enque(Packet*) = 0;
	virtual Packet* deque() = 0;

	void enque(Queue&, Packet*);
	Packet* deque(Queue&, int n);
	void drop(Queue&, int n);

	int busy_;		/* true "while packet is being transmitted" */
	Node* neighbor_;	/* next hop */

	double bandwidth_;	/* bandwidth of underlying link (bits/sec) */
	double delay_;		/* line latency */

	double busytime_;	/* total time link busy */

	inline void check_class(int c) {
		if (c >= ncs_)
			grow(c);
	}
	int ncs_;
	classStat* cs_;
	void grow(int n);
	Integrator ib_;		/* bytes */
	Integrator ip_;		/* packets */
	Integrator iu_;		/* utilization */

	void plot();
	Trace* qtrace_;
	Event intr_;
};

inline void Link::enque(Queue& q, Packet* p)
{
	++qnp_;
	qnb_ += p->size_;
	q.enque(p);
	update(p);
	if (trace_ != 0)
		trace_->enque(src_, dst_, p);
	if (qtrace_ != 0)
		plot();
}

inline Packet* Link::deque(Queue& q, int n)
{
	Packet* p = q.deque(n);
	qnb_ -= p->size_;
	--qnp_;
	update(p);
	if (trace_ != 0)
		trace_->deque(src_, dst_, p);
	if (qtrace_ != 0)
		plot();
	return (p);
}
	
inline void Link::drop(Queue& q, int n)
{
	/*XXX stats from qd.c */
	Packet* p = q.deque(n);
	qnb_ -= p->size_;
	--qnp_;
	update_drop(p);
	if (trace_ != 0)
		trace_->drop(src_, dst_, p);
	if (qtrace_ != 0)
		plot();
	Packet::free(p);
}
	
class LinkMatcher {
 public:
	static Link* lookup(const char* type);
 protected:
	LinkMatcher();
 private:
	virtual Link* match(const char* type) = 0;
	static LinkMatcher* all_;
	LinkMatcher* next_;
};

#endif
