/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: node.cc,v 1.6 95/09/29 15:09:41 mccanne Exp $ (LBL)";
#endif

#include <stdio.h>
#include <stdlib.h>

#include "packet.h"
#include "node.h"
#include "link.h"
#include "agent.h"

static class NodeCommand : public CommandHandler {
public:
	NodeCommand(const char* name) : CommandHandler(name) { }
        virtual int command(int argc, const char*const* argv);
} node_cmd("node");

int NodeCommand::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if (argc == 2) {
		Node* p = new Node;
		tcl.result(p->proc());
		return (TCL_OK);
	}
	tcl.result("ns node: arg mismatch");
	return (TCL_ERROR);
}

int Node::cnt_ = 0;

Node::Node() : Object('n'), nport_(0), route_(0), links_(0)
{
	addr_ = cnt_++;
	maxport_ = 16;
	demux_ = new Agent*[maxport_];
	memset(mrtab_, 0, sizeof(mrtab_));
	ngroup_ = 0;
	group_ = 0;
}

Node::~Node()
{
	delete route_;
	delete demux_;
	for (int i = ngroup_; --i >= 0; )
		free(group_[i]);
	delete group_;
}

void Node::free(AgentList* p)
{
	while (p != 0) {
		AgentList* n = p->next;
		delete p;
		p = n;
	}
}

void Node::add_link(Link* p)
{
	p->next_ = links_;
	links_ = p;
	links_->src(addr_);
}

void Node::setroute(int dst, Link* p)
{
	if (route_ == 0) {
		route_ = new Link*[cnt_];
		memset(route_, 0, cnt_* sizeof(Link*));
	}
	route_[dst] = p;
}

int Node::add_agent(Agent* p)
{
	int port = nport_++;
	if (port > maxport_) {
		int n = maxport_ * 2;
		Agent** demux = new Agent*[n];
		for (int i = 0; i < maxport_; ++i)
			demux[i] = demux_[i];
		delete demux_;
		demux_ = demux;
		maxport_ = n;
	}
	demux_[port] = p;
	return (port);
}

void Node::join_group(Agent* agent, int group)
{
	int n, i;

	group &=~ MCASTADDR;
	if (group_ == 0) {
		for (n = 16; n < group; n <<= 1)
			;
		group_ = new AgentList*[n];
		memset(group_, 0, n * sizeof(AgentList*));
		ngroup_ = n;
	} else if (group >= ngroup_) {
		for (n = ngroup_ << 1; n < group; n <<= 1)
			;
		AgentList** p = new AgentList*[n];
		for (i = 0; i < ngroup_; ++i)
			p[i] = group_[i];
		for (; i < n; ++i)
			p[i] = 0;
		delete group_;
		group_ = p;
	}
	AgentList* al = new AgentList;
	al->agent = agent;
	al->next = group_[group];
	group_[group] = al;
}

/*
 * Compute the reverse-path forwarding links.
 * Look through all the oubound links.  A next-hop node
 * that has a route to the source back through us, means
 * that it's on the reverse-path forwarding distribution.
 */
void Node::compute_mroute(mrt* mr)
{
	for (Link* p = links_; p != 0; p = p->next_) {
		Node* dst = p->neighbor();
		Link* back = dst->route_[mr->src];
		if (back != 0 && back->dst() == addr_) {
			mlink* ml = new mlink;
			ml->prune = 0;
			ml->link = p;
			ml->next = mr->links;
			mr->links = ml;
		}
	}
}

mrt* Node::mlookup(int src, int grp)
{
	mrt* p;

	int h = MRHASH(src, grp);
	for (p = mrtab_[h]; p != 0; p = p->next)
		if (p->src == src && p->grp == grp)
			return (p);
	
	p = new mrt;
	p->src = src;
	p->grp = grp;
	/*XXX*/
	p->links = 0;
	p->agents = 0;
	p->next = mrtab_[h];
	mrtab_[h] = p;

	compute_mroute(p);

	return (p);
}

/*
 * Nodes only ever see "packet" events, which come either
 * from an incoming link or a local agent (i.e., packet source).
 */
void Node::handle(Event* e)
{
	if (route_ == 0) {
		fprintf(stderr, "ns: can't run without routes\n");
		exit(1);
	}
	Packet* p = (Packet*)e;
	if (p->dst_ & MCASTADDR) {
		/*
		 * This is a multicast packet.  Forward it to all
		 * nodes along the reverse-path forwarding tree,
		 * and deliver to all local agents that are members
		 * of the corresponding multicast group.
		 */
		int g = p->dst_ &~ MCASTADDR;
		AgentList* local = (g < ngroup_) ? group_[g] : 0;
		mrt* m = mlookup(p->src_, g);
		mlink* ml = m->links;
		/*
		 * single hop special case - don't make any copies
		 */
		if (ml != 0 && ml->next == 0 && local == 0 && !ml->prune) {
			ml->link->send(p);
			return;
		}
		for (; ml != 0; ml = ml->next) {
			if (!ml->prune)
				ml->link->send(p->copy());
		}
		/* we use only the group addr (not port) for local demux */
		for (; local != 0; local = local->next)
			local->agent->recv(p->copy());
		Packet::free(p);
		return;
	}
	if (p->dst_ == addr_) {
		/* packet destined to this node, lookup agent */
		int port = p->dport_;
		if (port > nport_)
			abort();
		demux_[port]->recv(p);
		return;
	}
	Link* link = route_[p->dst_];
	if (link == 0) {
		/*XXX drop packet and bump error counter */
		fprintf(stderr, "no route for packet\n");
		exit(1);
	}
	link->send(p);
}

/*
 * $node addr
 * $node agent $port
 */
int Node::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if (argc == 2) {
		if (strcmp(argv[1], "addr") == 0) {
			tcl.resultf("%d", addr_);
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(argv[1], "agent") == 0) {
			int port = atoi(argv[2]);
			if (port < maxport_) {
				Agent* a = demux_[port];
				if (a != 0)
					tcl.result(a->proc());
			}
			return (TCL_OK);
		}
	}

	tcl.resultf("ns: no such node command: %s", argv[1]);
	return (TCL_ERROR);
}
